% basecnv.sl, a set of any base convertion routines.
%
% version 0.99
% By Dino Leonardo Sangoi.
% (I wrote this the first time trying to enhance char_table.sl, by
% Guenter Milde, so credits woes to him).

implements("BaseCnv");
% Uhm, maybe the user prefers lowercase chars. Should I make it configurable ?
% (I don't want to put yet another parameter to basecnv_itoa()).
static variable BaseChar = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";

%!%+
%\function{basecnv_pad}
%\synopsis{Pad a string to width filling with filler}
%\usage{String_Type basecnv_pad(String_Type str, Integer_Type width, String_Type filler)}
%\description
%   basecnv_pad() pads the string \var{str} with the string \var{filler} up to a \var{width} length
%   If \var{width} is a negative number, the string will be padded to left, otherwise it
%   will be padded to right. If the resulting string is longer that \var{width},
%   it's returned as is, i.e. no truncation is performed.
%\notes
%   If the string \var{filler} is longer that 1 char, bad things happen
%\seealso{basecnv_itoa, basecnv_atoi}
%!%-
public define basecnv_pad(str, width, filler)
{
	variable l = abs(width) - strlen(str);

	if (width == 0)
		return str;

	""; loop (l) () + filler;
	if (width > 0)
		return () + str;
	else
		return str + ();
}

%!%+
%\function{basecnv_itoa}
%\synopsis{Convert an integer to a string, according to the given base}
%\usage{String_Type basecnv_itoa(Integer_Type number, Integer_Type base, [Integer_Type width, Char_Type filler])}
%\description
%   basecnv_itoa() returns a string rappresentation of integer \var{number} expressed
%   in base \var{base}. If optional parameters \var{width} is present, the string is padded
%   using the basecnv_pad() function.
%
%\notes
%   This is a pure slang implementation, so converting to well known (and
%   fixed) base may be performed much faster using sprintf().
%\seealso{basecnv_pad, basecnv_atoi}
%!%-
public define basecnv_itoa()
{
	variable number, base, width = 0, filler = " ";
        variable j, s = "";
	variable l;

	if (_NARGS < 2 or _NARGS > 4)
		error("basecnv_itoa: not enough arguments");
	if (_NARGS > 3)
		filler = char(());
	if (_NARGS > 2)
		width = ();
	base = ();
	number = ();

        while (number) {
                j = number mod base;
                s = char(BaseChar[j]) + s;
                number = (number - j) / base;
        }
        if (s == "")
                s = "0";

	return basecnv_pad(s, width, filler);
}

%!%+
%\function{basecnv_atoi}
%\synopsis{Convert a string to an integer, according to the given base}
%\usage{Integer_Type basecnv_atoi(String_Type str, Integer_Type base)}
%\description
%   basecnv_atoi() converts the string str, expressed in base base, to an
%   integer.
%\notes
%   Wouldn't be nice if slang exports a strtol like function? (Hint, hint :))
%
%\seealso{basecnv_pad, basecnv_itoa}
%!%-
public define basecnv_atoi(str, base)
{
        variable rsign = 1, v, r = 0, i, c;

	str = strtrim_beg(str);
	if (str[0] == '-')
		rsign = -1;
	if (rsign == -1 or str[0] == '+')
		str = str[[1:]];
	str = strup(str);
	for (i=0; str[i] > ' '; i++) {
                c = str[i];
                if (c >= 'A')
                        v = c - 'A' + 10;
                else v = c - '0';
                if ((v < 0) or (v >= base))
			verror("basecnv_atoi: char '%c' out of base <%d> range.", c, base);
                r = r * base + v;
        }
        return rsign*r;
}

provide("basecnv");
