% Functions and variables for (un-)commenting lines and regions
% taken from ide.sl
% modified by GM <g.milde@physik.tu-dresden.de>
% modified by JED
% Modified by Saurus to allow "commenting comments".

static variable Comment_Data = Assoc_Type[Struct_Type];

%!%+
%\function{set_comment_info}
%\synopsis{Set comment information for a mode}
%\usage{set_comment_info ([mode,] cbeg, cend, flags)}
%\description
% This function sets comment information for a specified mode. If the
% optional mode argument is not present, the current mode will be used.  The
% other 3 required arguments represent the comment start string (\var{cbeg}),
% the comment end string (\var{cend}), and an integer flags argument that
% indications how these strings are to be used by the \var{comment_region}
% function.  In particular, \var{flags} is a bitmapped integer whose bits
% have the following meaning:
%#v+
%     0x01  :  Comments will start at column defined by the region start,
%               otherwise comments will be indented to the level of the
%               first line in the region.
%     0x02  :  Lines in the region will be padded out to the same column.
%     0x04  :  Blank lines will be commented.
%     0x08  :  Old comments will be converted to strings found in the
%              comment info associated with _{mode}bak mode.
%#v-
%\seealso{get_comment_info, comment_region, uncomment_region, comment_line, uncomment_line, comment_region_or_line, uncomment_region_or_line}
%!%-
public define set_comment_info (cbeg, cend, flags)
{
   if (_NARGS == 3)
     get_mode_name ();
   variable mode = ();

   variable s = struct
     {
        cbeg, cend, flags
     };
   s.cbeg = cbeg;
   s.cend = cend;
   s.flags = flags;

   Comment_Data[mode] = s;
}

%!%+
%\function{get_comment_info}
%\synopsis{Get comment information according to mode} }
%\usage{Struct_Type = get_comment_info ( [mode] ); }
%\description
% Retrieves the comment information according to the optional \var{mode}
% argument, or for the present mode it \var{mode} is not present.
% Every mode that wants to use this function should provide comment information
% using the \var{set_comment_info} function.
%
% The returned value is a structure with the following fields:
%#v+
%     cbeg       % begin comment string
%     cend       % end comment string
%     flags      % flags
%#v-
% If comment information does not exist for the mode, then \var{NULL} will
% be returned.
%\seealso{set_comment_info, comment_region, uncomment_region, comment_line, uncomment_line, comment_region_or_line, uncomment_region_or_line}
%!%-
public define get_comment_info ()
{
   !if (_NARGS)
     get_mode_name ();

   variable modename = ();

   !if (assoc_key_exists (Comment_Data, modename))
     return NULL;

   return Comment_Data[modename];
}

% search for already present comments, and replace those with comments strings
% from the _{mode}bak fake mode.
% if redo != 0, put back good comments.
static define seek_and_destroy_old_comments(redo)
{
   variable bakmode = "_" + get_mode_name() + "bak";
   variable newi = get_comment_info();
   variable bak = get_comment_info(bakmode);
   variable new = struct
     {
        cbeg, cend
     };

   if ((newi == NULL) or (bak == NULL))
     return;

   % I Suppose that white spaces have no meaning, and are here only for aesthetic purposes
   new.cbeg = strtrim(newi.cbeg);
   new.cend = strtrim(newi.cend);

   if (redo)
     {
        variable t = new;
        new = bak;
        bak = t;
     }

   replace(new.cbeg, bak.cbeg);
   replace(new.cend, bak.cend);
}

static define _get_comment_info ()
{
   variable m = get_mode_name ();
   variable s = get_comment_info (m);
   if (s == NULL)
     verror("No comment strings defined for %s mode", m);
   return s;
}

static define compute_max_column (beg_mark, end_mark)
{
   variable max_column;

   goto_user_mark (end_mark);
   max_column = what_column ();
   goto_user_mark (beg_mark);

   forever
     {
        eol ();
        if (create_user_mark () >= end_mark)
          break;
        bskip_white ();
        if (what_column () > max_column)
          max_column = what_column ();
        go_down_1 ();
     }
   return max_column;
}

%!%+
%\function{comment_region}
%\synopsis{Comment a region}
%\usage{comment_region ()}
%\description
%   This function will comment a region using the comment info registered
%   to the mode associated with this buffer via the \var{set_comment_info}
%   function.
%\notes
%   Bind to a key.
%\seealso{set_comment_info, get_comment_info, uncomment_region, comment_line, uncomment_line, comment_region_or_line, uncomment_region_or_line}
%!%-
public define comment_region ()
{
   check_region (1);

   ERROR_BLOCK
     {
        pop_spot ();
        pop_mark_0 ();
     }

   variable info = _get_comment_info ();
   variable cbeg = info.cbeg;
   variable cend = info.cend;
   variable flags = info.flags;

   variable end_mark = create_user_mark ();
   exchange_point_and_mark ();
   variable beg_mark = create_user_mark ();
   variable indent_col = 1;

   if (beg_mark == end_mark)
     {
        pop_spot ();
        pop_mark_0 ();
        insert (cbeg);
        insert (cend);
        return;
     }

   if (flags & 0x01)
     {
        indent_col = what_column ();
        bskip_white ();
        !if (bolp())
          indent_col = 1;
     }
   else
     {
        skip_chars ("\n\t ");
        bskip_white ();
        if (bolp ())
          {
             skip_white ();
             if (create_user_mark () < end_mark)
               {
                  indent_col = what_column ();
                  move_user_mark (beg_mark);
               }
          }
     }

   goto_user_mark (end_mark);
   bskip_chars ("\n\t ");

   variable ok_to_pad_end = 0;
   if (create_user_mark () > beg_mark)
     {
        move_user_mark (end_mark);
        pop_mark_0 ();
        push_mark ();
        skip_white ();
        ok_to_pad_end = eolp ();
     }

   variable max_column = 0;
   if ((flags & 0x02) and strlen (cend))
     max_column = compute_max_column (beg_mark, end_mark) + strlen (cbeg);


   goto_user_mark (beg_mark);
   insert (cbeg);
   bol ();
   narrow_to_region ();

   % Check old comment handling
   if (flags & 0x08)
     {
        % Uhmmm, the first cbeg was already inserted, search it.
        if (fsearch(cbeg) > 0)
          {
             go_right_1;
             seek_and_destroy_old_comments(0);
          }
     }

   ERROR_BLOCK
     {
        widen_region ();
        pop_spot ();
     }
   eol ();

   if (max_column)
     {
        trim ();
        if (what_column () < max_column)
          goto_column (max_column);
     }
   insert (cend);

   variable comment_blank_lines = flags & 0x04;

   forever
     {
        if (comment_blank_lines)
          {
             !if (down_1 ())
               break;
             skip_white ();
             if (eolp ())
               goto_column (indent_col);
          }
        else
          {
             skip_chars ("\n\t ");
             if (eobp ())
               break;
          }

        if (what_column () >= indent_col)
          goto_column (indent_col);

        insert (cbeg);
        eol ();
        if (max_column)
          {
             if ((create_user_mark () < end_mark ())
                 or ok_to_pad_end)
               {
                  trim ();
                  if (what_column () < max_column)
                    goto_column (max_column);
               }
          }
        insert (cend);
     }
   widen_region ();
   pop_spot ();
}

%!%+
%\function{uncomment_region}
%\synopsis{Uncomment a region}
%\usage{Void uncomment_region ()}
%\description
%   This function uncomments a region previously commented with \var{comment_region}
%   or \var{comment_region_or_line}. It may also work for hand commented regions,
%   but depends on the commenting style used.
%\notes
%   Bind to a key.
%\seealso{set_comment_info, get_comment_info, comment_region, comment_line, uncomment_line, comment_region_or_line, uncomment_region_or_line}
%!%-
public define uncomment_region ()
{
   check_region (1);
   variable end_mark = create_user_mark ();
   exchange_point_and_mark ();
   narrow_to_region ();

   ERROR_BLOCK
     {
        widen_region ();
        pop_spot ();
     }

   variable info = _get_comment_info ();
   variable cbeg = info.cbeg;
   variable cend = info.cend;
   variable do_trim = info.flags & 0x02;

   variable len_cbeg = strlen (cbeg);
   variable len_cend = strlen (cend);

   if (info.flags & 0x08)
     seek_and_destroy_old_comments(1);

   do
     {
        !if (ffind (cbeg))
          continue;

        deln (len_cbeg);
        eol ();
        !if (bfind (cend))
          continue;

        deln (len_cend);
        if (do_trim)
          {
             skip_white ();
             if (eolp ())
               trim ();
          }
     }
   while (down_1 ());
   widen_region ();
   pop_spot ();
}

%!%+
%\function{comment_line}
%\synopsis{Comment current line}
%\usage{Void comment_line ()}
%\description
%   This function will comment the current line using the comment info
%   registered to the mode associated with this buffer via the
%   \var{set_comment_info} function.
%\notes
%   Bind to a key.
%\seealso{set_comment_info, get_comment_info, comment_region, uncomment_region, uncomment_line, comment_region_or_line, uncomment_region_or_line}
%!%-
public define comment_line ()
{
   ERROR_BLOCK { pop_spot (); }
   push_spot_bol ();
   push_mark_eol ();
   comment_region ();
   pop_spot ();
}

%!%+
%\function{uncomment_line}
%\synopsis{Uncomment current line}
%\usage{Void uncomment_line ()}
%\description
%   This function uncomments the current line previously commented with
%   \var{comment_line} or \var{comment_region_or_line}. It may also work for hand
%   commented lines, but depends on the commenting style used. obviously you
%   can also uncomment a single line from a commented region.
%\notes
%   Bind to a key.
%\seealso{set_comment_info, get_comment_info, comment_region, uncomment_region, comment_line, comment_region_or_line, uncomment_region_or_line}
%!%-
public define uncomment_line ()
{
   ERROR_BLOCK { pop_spot (); }
   push_spot_bol ();
   push_mark_eol ();
   uncomment_region ();
   pop_spot ();
}

%!%+
%\function{comment_region_or_line}
%\synopsis{Comment a region or current line}
%\usage{comment_region ()}
%\description
%   This function looks at the current region (or line, if mark is not set).
%   If looks like a comment was already set, uncomments it. otherwise the
%   region or line are commented. It uses the comment info registered to the
%   mode associated with this buffer via the \var{set_comment_info} function.
%\notes
%   Bind to a key.
%\seealso{set_comment_info, get_comment_info, comment_region, uncomment_region, comment_line, uncomment_line, uncomment_region_or_line}
%!%-
public define comment_region_or_line ()
{
   if (-1 != prefix_argument (-1))
     {
        uncomment_region_or_line ();
        return;
     }

   if (markp ())
     return comment_region ();

   variable s = _get_comment_info ();
   push_spot ();
   bol_skip_white ();
   if (looking_at (s.cbeg))
     {
        pop_spot ();
        return uncomment_line ();
     }
   pop_spot ();
   comment_line ();
}


%!%+
%\function{uncomment_region_or_line}
%\synopsis{Uncomment a region or current line}
%\usage{uncomment_region_or_line ()}
%\description
%   This functions uncomments a region if mark is set, otherwise it uncomments
%   the current line.
%\notes
%   Bind to a key.
%\seealso{set_comment_info, get_comment_info, comment_region, uncomment_region, comment_line, uncomment_line, comment_region_or_line}
%!%-
public define uncomment_region_or_line ()
{
   if (markp ())
     uncomment_region ();
   else
     uncomment_line ();
}

% for some modes we have comment string definitons here
set_comment_info ("html", "<!-- ", " -->", 8);
set_comment_info ("_htmlbak", "<!!-!- ", " -!-!>", 0);
set_comment_info ("sgml", "<!-- ", " -->", 8);
set_comment_info ("_sgmlbak", "<!!-!- ", " -!-!>", 0);
set_comment_info ("docbook", "<!-- ", " -->", 8);
set_comment_info ("_docbookbak", "<!!-!- ", " -!-!>", 0);
set_comment_info ("C", "/* ", " */", 8);
set_comment_info ("_Cbak", "/.* ", " *./", 0);
set_comment_info ("SLang", "% ", "", 0);
set_comment_info ("TeX", "% ", "", 0);
set_comment_info ("LaTeX", "% ", "", 0);
set_comment_info ("SH", "# ", "", 0);
set_comment_info ("matlab", "# ", "", 0);
set_comment_info ("perl", "# ", "", 0);
set_comment_info ("Fortran", "C ", "", 0);
set_comment_info ("TPas", "{ ", " }", 0);
set_comment_info ("PHP", "// ", "", 0);

provide ("comments");

